# Copyright (C) 2012-2013  Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import pywbem
import collections

_RValue = collections.namedtuple('RValue', ["rval", "rparams", "errorstr"])

def raise_or_dump_exception(e):
    if not LmiBaseClient._get_use_exceptions():
        return
    (et, ei, tb) = sys.exc_info()
    raise et, ei, tb

class LmiReturnValue(_RValue):
    def __getattr__(self, name):
        if not name in self._properties:
            raise AttributeError(name)
        return self._properties[name]

    def properties(self):
        return self._properties.keys()

    @property
    def use_display_hook(self):
        return self._properties["use_display_hook"]

    @use_display_hook.setter
    def use_display_hook(self, value):
        self._properties["use_display_hook"] = value

    @property
    def instance_classname(self):
        return self._properties["instance"].classname if self._properties["instance"] else ""

    def __new__(self, **kwargs):
        properties = { "use_display_hook" : False }
        keys_to_pop = []
        for (key, value) in kwargs.iteritems():
            if key in ("rval", "rparams", "errorstr"):
                continue
            properties[key] = value
            keys_to_pop.append(key)
        for key in keys_to_pop:
            kwargs.pop(key)
        rval = _RValue.__new__(self, **kwargs)
        rval._properties = properties
        return rval

class LmiBaseClient(object):
    QUERY_LANG_CQL = "CQL"
    QUERY_LANG_WQL = "WQL"
    use_exceptions_flag = False

    def __init__(self, hostname, username = "", password = ""):
        self._hostname = hostname
        self._username = username
        self._password = password
        if not self._hostname.startswith("http://") and \
            not self._hostname.startswith("https://"):
            self._hostname = "https://" + self._hostname
        self._cliconn = pywbem.WBEMConnection(self._hostname,
            (self._username, self._password))

    def _get_instance_names(self, klass, **kwargs):
        try:
            inst_name_list = self._cliconn.EnumerateInstanceNames(klass, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = None, rparams = {}, errorstr = e.args[1])
        except pywbem.cim_http.AuthError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = None, rparams = {}, errorstr = e.args[0])
        return LmiReturnValue(rval = inst_name_list, rparams = {}, errorstr = "")

    def _get_instance(self, path, **kwargs):
        inst = self._cliconn.GetInstance(path, **kwargs)
        if not inst:
            return LmiReturnValue(rval = None, rparams = {}, errorstr = "No such class instance")
        return LmiReturnValue(rval = inst, rparams = {}, errorstr = "")

    def _get_instances(self, klass, **kwargs):
        filter_value = ""
        filter_key = "Name"
        if "key" in kwargs:
            filter_key = kwargs["key"]
            kwargs.pop("key")
        elif "Key" in kwargs:
            filter_key = kwargs["Key"]
            kwargs.pop("Key")
        if "value" in kwargs:
            filter_value = kwargs["value"]
            kwargs.pop("value")
        if "Value" in kwargs:
            filter_value = kwargs["Value"]
            kwargs.pop("Value")
        (inst_name_list, rparams, errorstr) = self._get_instance_names(klass, **kwargs)
        if not inst_name_list:
            return LmiReturnValue(rval = None, rparams = {}, errorstr = errorstr)
        if filter_value:
            inst_name_list = filter(lambda n: not filter_key in n or n[filter_key] == filter_value, inst_name_list)
        inst_list = []
        for i in inst_name_list:
            try:
                inst = self._cliconn.GetInstance(InstanceName=i, LocalOnly = False)
            except pywbem.cim_operations.CIMError, e:
                raise_or_dump_exception(e)
                # we can get this error for eg. Linux_UnixProcess instances
                # when, there the process exited earlier than this method
                if e.args[0] == pywbem.CIM_ERR_NOT_FOUND:
                    # what is right, skip non existing object, or drop the whole list?
                    continue
                else:
                    return LmiReturnValue(rval = None, rparams = {}, errorstr = e.args[1])
            if not filter_value or filter_key in inst and inst[filter_key] == filter_value:
                inst_list.append(inst)
        if not inst_list:
            return LmiReturnValue(rval = None, rparams = {}, errorstr = "Not found" if not "ErrorMsg" in kwargs else kwargs["ErrorMsg"])
        return LmiReturnValue(rval = inst_list, rparams = {}, errorstr = "")

    def _delete_instance(self, inst):
        try:
            self._cliconn.DeleteInstance(inst.path)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = False, rparams = {}, errorstr = e.args[1])
        return LmiReturnValue(rval = True, rparams = {}, errorstr = "")

    def _get_class_names(self, namespace = None, **kwargs):
        try:
            class_name_list = self._cliconn.EnumerateClassNames(namespace, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = None, rparams = {}, errorstr = e.args[1])
        except pywbem.cim_http.AuthError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = None, rparams = {}, errorstr = e.args[0])
        return LmiReturnValue(rval = class_name_list, rparams = {}, errorstr = "")

    def _get_classes(self, namespace = None, **kwargs):
        try:
            class_list = self._cliconn.EnumerateClasses(namespace, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = None, rparams = {}, errorstr = e.args[1])
        except pywbem.cim_http.AuthError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = None, rparams = {}, errorstr = e.args[0])
        return LmiReturnValue(rval = class_list, rparams = {}, errorstr = "")

    def _get_class(self, class_name, namespace = None, **kwargs):
        try:
            klass = self._cliconn.GetClass(class_name, namespace, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = None, rparams = {}, errorstr = e.args[1])
        except pywbem.cim_http.AuthError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = None, rparams = {}, errorstr = e.args[0])
        return LmiReturnValue(rval = klass, rparams = {}, errorstr = "")

    def _call_method_raw(self, instance, method, **params):
        try:
            (rval, rparams) = self._cliconn.InvokeMethod(method, instance.path, **params)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = -1, rparams = {}, errorstr = e.args[1] + ": '" + method + "'")
        return LmiReturnValue(rval = rval, rparams = rparams, errorstr = "")

    def _call_method(self, instance, method, **params):
        (rval, rparams, errorstr) = self._call_method_raw(instance, method, **params)
        return LmiReturnValue(rval = rval == 0, rparams = rparams, errorstr = errorstr)

    def _get_associators(self, instance, **params):
        try:
            return self._cliconn.Associators(instance.path, **params)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return []

    def _get_references(self, instance, **params):
        try:
            return self._cliconn.References(instance.path, **params)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return []

    def _modify_instance(self, instance, **params):
        try:
            self._cliconn.ModifyInstance(instance, **params)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = e.args[0], rparams = {}, errorstr = e.args[1])
        return LmiReturnValue(rval = 0, rparams = {}, errorstr = "")

    def _delete_instance(self, instance, **params):
        try:
            self._cliconn.DeleteInstance(instance, **params)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = e.args[0], rparams = {}, errorstr = e.args[1])
        return LmiReturnValue(rval = 0, rparams = {}, errorstr = "")

    def _exec_query(self, query_lang, query, namespace = None):
        try:
            inst_list = self._cliconn.ExecQuery(query_lang, query, namespace)
        except pywbem.cim_operations.CIMError, e:
            raise_or_dump_exception(e)
            return LmiReturnValue(rval = None, rparams = {}, errorstr = e.args[1])
        return LmiReturnValue(rval = inst_list, rparams = {}, errorstr = "")

    @staticmethod
    def _set_use_exceptions(use = True):
        LmiBaseClient.use_exceptions_flag = use

    @staticmethod
    def _get_use_exceptions():
        return LmiBaseClient.use_exceptions_flag

    @property
    def username(self):
        return self._username

    @property
    def hostname(self):
        return self._hostname
